/*
 * ProFTPD: mod_auth_external -- auth with call external command
 *
 * Copyright (c) 2007 by Sergey V. Sokolov <sokol@tzavolga.net> / All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#include "conf.h"
#include "privs.h"
#include <stdlib.h>
#include <unistd.h>

#ifdef PR_USE_SHADOW
# include <shadow.h>
#endif

#ifdef HAVE_SYS_SECURITY_H
# include <sys/security.h>
#endif

#if defined(HAVE_CRYPT_H) && !defined(AIX4) && !defined(AIX5)
# include <crypt.h>
#endif

#ifdef SETGRENT_VOID
# define RETSETGRENTTYPE void
#else
# define RETSETGRENTTYPE int
#endif

// Directives
#define DIRECTIVE_AUTHEXTCMD  "AuthExternalCmd"
#define DIRECTIVE_AUTHEXTEXEC "AuthExecAsRoot"

// Exit status codes
#define EXT_AUTH_NOPWD		1
#define EXT_AUTH_BADPWD 	2
#define EXT_AUTH_AGEPWD 	3
#define EXT_AUTH_DISABLEPWD 	4
#define EXT_ERROR_BADCALL 	5
#define EXT_ERROR_BADCONF 	6
#define EXT_ERROR_BADDBCON 	7

// Structs
typedef union idauth
{
	uid_t uid;
	gid_t gid;
} idauth_t;

typedef struct _idmap
{
	struct _idmap *next, *prev;

	// This is a union because different OSs may give different types to
	// UIDs
	// and GIDs.  This presents a far more portable way to deal with this
	// reality.
	idauth_t id;

	// User or group name
	char *name;
} idmap_t;

// Tables
#define HASH_TABLE_SIZE 128
static xaset_t *uid_table[HASH_TABLE_SIZE];
static xaset_t *gid_table[HASH_TABLE_SIZE];

// Path to external command
config_rec *ae_external_cmd = NULL;
unsigned char *ae_exec_as_root = NULL;

// Initialization
static int authext_sess_init(void)
{
	ae_external_cmd = find_config(main_server->conf, CONF_PARAM, DIRECTIVE_AUTHEXTCMD, FALSE);
	ae_exec_as_root = get_param_ptr(main_server->conf, DIRECTIVE_AUTHEXTEXEC, FALSE);
	return 0;
}

// 
// Utils
// 

inline static int _compare_uid(idmap_t * m1, idmap_t * m2)
{
	if (m1->id.uid < m2->id.uid)
		return -1;

	if (m1->id.uid > m2->id.uid)
		return 1;

	return 0;
}

inline static int _compare_gid(idmap_t * m1, idmap_t * m2)
{
	if (m1->id.gid < m2->id.gid)
		return -1;

	if (m1->id.gid > m2->id.gid)
		return 1;

	return 0;
}

inline static int _compare_id(xaset_t ** table, idauth_t id, idauth_t idcomp)
{
	if (table == uid_table)
		return id.uid == idcomp.uid;
	else
		return id.gid == idcomp.gid;
}

static idmap_t *_auth_lookup_id(xaset_t ** id_table, idauth_t id)
{
	int hash = ((id_table == uid_table) ? id.uid : id.gid) % HASH_TABLE_SIZE;
	idmap_t *m;

	if (!id_table[hash])
		id_table[hash] = xaset_create(session.pool, (id_table == uid_table) ? (XASET_COMPARE) _compare_uid : (XASET_COMPARE) _compare_gid);

	for (m = (idmap_t *) id_table[hash]->xas_list; m; m = m->next) {
		if (_compare_id(id_table, m->id, id))
			break;
	}

	if (!m || !_compare_id(id_table, m->id, id)) {
		// Isn't in the table
		m = (idmap_t *) pcalloc(id_table[hash]->pool, sizeof(idmap_t));

		if (id_table == uid_table)
			m->id.uid = id.uid;
		else
			m->id.gid = id.gid;

		xaset_insert_sort(id_table[hash], (xasetmember_t *) m, FALSE);
	}

	return m;
}

// Execute external programm
static int exec_external(char *cmd, char *const argv[], char *const envp[])
{
	int pid, status;
	if ((pid = fork()) < 0) {
		// Fork failed
		return -4;
	} else if (pid == 0) {
		execve(cmd, argv, envp);
		exit(-1);
	} else {
		waitpid(pid, &status, 0);
		return WIFEXITED(status) ? WEXITSTATUS(status) : -2;
	}
}

// Setpwent implementation
MODRET authext_setpwent(cmd_rec * cmd)
{
	setpwent();
	return DECLINED(cmd);
}

// Endpwent implementation
MODRET authext_endpwent(cmd_rec * cmd)
{
	endpwent();
	return DECLINED(cmd);
}

// Setgrent implementation
MODRET authext_setgrent(cmd_rec * cmd)
{
	setgrent();
	return DECLINED(cmd);
}

// Endgrent implementation
MODRET authext_endgrent(cmd_rec * cmd)
{
	endgrent();
	return DECLINED(cmd);
}

// Getgrent implementation
MODRET authext_getgrent(cmd_rec * cmd)
{
	struct group *gr;
	gr = getgrent();
	return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
}

// Getpwent implementation
MODRET authext_getpwent(cmd_rec * cmd)
{
	struct passwd *pw;
	pw = getpwent();
	return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
}

// Getpwuid implementation
MODRET authext_getpwuid(cmd_rec * cmd)
{
	struct passwd *pw;
	uid_t uid;

	uid = *((uid_t *) cmd->argv[0]);
	pw = getpwuid(uid);

	return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
}

// Getpwnam implementation
MODRET authext_getpwnam(cmd_rec * cmd)
{
	struct passwd *pw;
	const char *name;

	name = cmd->argv[0];
	pw = getpwnam(name);

	return pw ? mod_create_data(cmd, pw) : DECLINED(cmd);
}

// Getgrnam implementation
MODRET authext_getgrnam(cmd_rec * cmd)
{
	struct group *gr;
	const char *name;

	name = cmd->argv[0];
	gr = getgrnam(name);

	return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
}

// Getgrgid implementation
MODRET authext_getgrgid(cmd_rec * cmd)
{
	struct group *gr;
	gid_t gid;

	gid = *((gid_t *) cmd->argv[0]);
	gr = getgrgid(gid);

	return gr ? mod_create_data(cmd, gr) : DECLINED(cmd);
}

// Auth
MODRET authext_auth(cmd_rec * cmd)
{
	time_t now = time(NULL), lstchg = -1, max = -1, inact = -1, disable = -1;
	const char *name = cmd->argv[0];
	const char *cleartxt_pass = cmd->argv[1];

	// Do not handle *pw* requests unless we can do so.
	if (!ae_external_cmd) {
		pr_log_pri(PR_LOG_ERR, "Directive '%s' not defined in config file!", DIRECTIVE_AUTHEXTCMD);
		return DECLINED(cmd);
	}

	const char *remote_ip = pr_netaddr_get_ipstr(session.c->remote_addr);

	if (pr_auth_check(cmd->tmp_pool, cleartxt_pass, name, cmd->argv[1]))
		return ERROR_INT(cmd, PR_AUTH_BADPWD);

	if (lstchg > 0 && max > 0 && inact > 0 && now > (lstchg + max + inact))
		return ERROR_INT(cmd, PR_AUTH_AGEPWD);

	if (disable > 0 && now > disable)
		return ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);

	// Call external command for auth
	char logcmd[255] = { '\0' };
	char *cmd_argv[5] = { ae_external_cmd->argv[0], name, cleartxt_pass, remote_ip, NULL };
	char *cmd_env[2] = { NULL };
	sprintf(logcmd, "%s '%s' '%s' '%s'", ae_external_cmd->argv[0], name, "(hidden)", remote_ip);

	if (ae_exec_as_root != NULL && *ae_exec_as_root == TRUE)
		PRIVS_ROOT 
	
	int retval = exec_external(ae_external_cmd->argv[0], cmd_argv, cmd_env);
	pr_log_pri(PR_LOG_INFO, "Execute external AuthCMD: %s", logcmd);
	pr_log_pri(PR_LOG_INFO, "AuthCMD exit code: %d", retval);

	if (ae_exec_as_root != NULL && *ae_exec_as_root == TRUE)
		PRIVS_RELINQUISH

	// Check exit code
	if (retval != 0) {
		if (retval < 0) {
			pr_log_pri(PR_LOG_ERR, "Can't execute AuthCMD: %s", logcmd);
			return DECLINED(cmd);
		} else if (retval > 0) {

			// Auth error
			switch (retval) {
			case EXT_AUTH_NOPWD:
				pr_log_pri(PR_LOG_INFO, "Can't find user '%s' (ip: %s) in passdb backend!", name, remote_ip);
				return ERROR_INT(cmd, PR_AUTH_NOPWD);

			case EXT_AUTH_BADPWD:
				pr_log_pri(PR_LOG_INFO, "Password mismatch for '%s' (ip: %s)!", name, remote_ip);
				return ERROR_INT(cmd, PR_AUTH_BADPWD);

			case EXT_AUTH_AGEPWD:
				pr_log_pri(PR_LOG_INFO, "Password expired for '%s' (ip: %s)!", name, remote_ip);
				return ERROR_INT(cmd, PR_AUTH_AGEPWD);

			case EXT_AUTH_DISABLEPWD:
				pr_log_pri(PR_LOG_INFO, "Account '%s' (ip: %s) deactivated!", name, remote_ip);
				return ERROR_INT(cmd, PR_AUTH_DISABLEDPWD);

			case EXT_ERROR_BADCALL:
				pr_log_pri(PR_LOG_ERR, "AuthExternalCMD bad call, params empty or not exists!");
				return DECLINED(cmd);

			case EXT_ERROR_BADCONF:
				pr_log_pri(PR_LOG_ERR, "Bad config for external auth cmd!");
				return DECLINED(cmd);

			case EXT_ERROR_BADDBCON:
				pr_log_pri(PR_LOG_ERR, "Bad connection to passdb backend!");
				return DECLINED(cmd);

			default:
				pr_log_pri(PR_LOG_ERR, "Uncatched exit code - %d!", retval);
				return ERROR_INT(cmd, PR_AUTH_BADPWD);
			}
		}
	}

	session.auth_mech = "mod_auth_external.c";
	return HANDLED(cmd);
}

// Pass check (do nothing, check in auth function)
MODRET authext_check(cmd_rec * cmd)
{
	session.auth_mech = "mod_auth_external.c";
	return HANDLED(cmd);
}

// Uid2name implementation
MODRET authext_uid2name(cmd_rec * cmd)
{
	idmap_t *m;
	idauth_t id;
	struct passwd *pw;

	id.uid = *((uid_t *) cmd->argv[0]);
	m = _auth_lookup_id(uid_table, id);

	if (!m->name) {
		/*
		 * Wasn't cached, so perform a lookup 
		 */
		pw = getpwuid(id.uid);

		if (pw) {
			m->name = pstrdup(session.pool ? session.pool : permanent_pool, pw->pw_name);
			return mod_create_data(cmd, m->name);
		}

		return DECLINED(cmd);
	}

	return mod_create_data(cmd, m->name);
}

// Gid2name implementation
MODRET authext_gid2name(cmd_rec * cmd)
{
	idmap_t *m;
	idauth_t id;
	struct group *gr;

	id.gid = *((gid_t *) cmd->argv[0]);

	m = _auth_lookup_id(gid_table, id);

	if (!m->name) {
		gr = getgrgid(id.gid);

		if (gr) {
			m->name = pstrdup(session.pool ? session.pool : permanent_pool, gr->gr_name);
			return mod_create_data(cmd, m->name);
		}

		return DECLINED(cmd);
	}

	return mod_create_data(cmd, m->name);
}

// Name2uid implementation
MODRET authext_name2uid(cmd_rec * cmd)
{
	struct passwd *pw;
	const char *name;

	name = cmd->argv[0];
	pw = getpwnam(name);

	return pw ? mod_create_data(cmd, (void *) &pw->pw_uid) : DECLINED(cmd);
}

// Name2gid implementation
MODRET authext_name2gid(cmd_rec * cmd)
{
	struct group *gr;
	const char *name;

	name = cmd->argv[0];
	gr = getgrnam(name);

	return gr ? mod_create_data(cmd, (void *) &gr->gr_gid) : DECLINED(cmd);
}

// Get user groups
// cmd->argv[0] = name
// cmd->argv[1] = (array_header **) group_ids
// cmd->argv[2] = (array_header **) group_names
MODRET authext_getgroups(cmd_rec * cmd)
{
	struct passwd *pw = NULL;
	struct group *gr = NULL;
	array_header *gids = NULL, *groups = NULL;
	char **gr_member = NULL, *name = NULL;

	// Function pointers for which lookup functions to use
	struct passwd *(*my_getpwnam) (const char *) = NULL;
	struct group *(*my_getgrgid) (gid_t) = NULL;
	struct group *(*my_getgrent) (void) = NULL;
	RETSETGRENTTYPE(*my_setgrent) (void) = NULL;

	// Play function pointer games
	my_getpwnam = getpwnam;
	my_getgrgid = getgrgid;
	my_getgrent = getgrent;
	my_setgrent = setgrent;

	name = (char *) cmd->argv[0];

	// Check for NULL values.
	if (cmd->argv[1])
		gids = (array_header *) cmd->argv[1];

	if (cmd->argv[2])
		groups = (array_header *) cmd->argv[2];

	// Retrieve the necessary info.
	if (!name || !(pw = my_getpwnam(name)))
		return DECLINED(cmd);

	// Populate the first group ID and name.
	if (gids)
		*((gid_t *) push_array(gids)) = pw->pw_gid;

	if (groups && (gr = my_getgrgid(pw->pw_gid)) != NULL)
		*((char **) push_array(groups)) = pstrdup(session.pool, gr->gr_name);

	my_setgrent();

	// This is where things get slow, expensive, and ugly.  Loop through
	// everything, checking to make sure we haven't already added it.
	while ((gr = my_getgrent()) != NULL && gr->gr_mem) {

		// Loop through each member name listed
		for (gr_member = gr->gr_mem; *gr_member; gr_member++) {

			// If it matches the given username...
			if (!strcmp(*gr_member, pw->pw_name)) {

				// Add the GID and name
				if (gids)
					*((gid_t *) push_array(gids)) = gr->gr_gid;

				if (groups && pw->pw_gid != gr->gr_gid)
					*((char **) push_array(groups)) = pstrdup(session.pool, gr->gr_name);
			}
		}
	}


	if (gids && gids->nelts > 0)
		return mod_create_data(cmd, (void *) &gids->nelts);
	else if (groups && groups->nelts > 0)
		return mod_create_data(cmd, (void *) &groups->nelts);

	return DECLINED(cmd);
}

// Set command
// Syntax: AuthExternalCmd <path>
MODRET set_authextcmd(cmd_rec * cmd)
{
	config_rec *c = NULL;

	CHECK_ARGS(cmd, 1);
	CHECK_CONF(cmd, CONF_ROOT | CONF_VIRTUAL | CONF_GLOBAL);

	c = add_config_param_str(DIRECTIVE_AUTHEXTCMD, 1, cmd->argv[1]);

	return HANDLED(cmd);
}

// Set exec as root
MODRET set_authexecasroot(cmd_rec * cmd)
{
	int bool = -1;
	config_rec *c = NULL;

	CHECK_ARGS(cmd, 1);
	CHECK_CONF(cmd, CONF_ROOT | CONF_VIRTUAL | CONF_GLOBAL);

	if ((bool = get_boolean(cmd, 1)) == -1)
		CONF_ERROR(cmd, "expected Boolean parameter");

	c = add_config_param(cmd->argv[0], 1, NULL);
	c->argv[0] = pcalloc(c->pool, sizeof(unsigned char));
	*((unsigned char *) c->argv[0]) = bool;

	return HANDLED(cmd);
}

// Module API tables
static conftable authext_conftab[] = {
	{DIRECTIVE_AUTHEXTCMD, set_authextcmd, NULL},
	{DIRECTIVE_AUTHEXTEXEC, set_authexecasroot, NULL},
	{NULL, NULL, NULL}
};

static authtable authext_authtab[] = {
	{0, "setpwent", authext_setpwent},
	{0, "endpwent", authext_endpwent},
	{0, "setgrent", authext_setgrent},
	{0, "endgrent", authext_endgrent},
	{0, "getpwent", authext_getpwent},
	{0, "getgrent", authext_getgrent},
	{0, "getpwnam", authext_getpwnam},
	{0, "getpwuid", authext_getpwuid},
	{0, "getgrnam", authext_getgrnam},
	{0, "getgrgid", authext_getgrgid},
	{0, "auth", authext_auth},
	{0, "check", authext_check},
	{0, "uid_name", authext_uid2name},
	{0, "gid_name", authext_gid2name},
	{0, "name_uid", authext_name2uid},
	{0, "name_gid", authext_name2gid},
	{0, "getgroups", authext_getgroups},
	{0, NULL, NULL}
};

// Module structure
module auth_external_module = {

	/*
	 * Always NULL 
	 */
	NULL, NULL,

	/*
	 * Module API version (2.0) 
	 */
	0x20,

	/*
	 * Module name 
	 */
	"auth_external",

	/*
	 * Module configuration directive handlers 
	 */
	authext_conftab,

	/*
	 * Module command handlers 
	 */
	NULL,

	/*
	 * Module authentication handlers (none in this case) 
	 */
	authext_authtab,

	/*
	 * Module initialization 
	 */
	NULL,

	/*
	 * Session initialization 
	 */
	authext_sess_init
};
